#include "activation.hpp"

namespace activation
{
ReluActivation        relu;
IdentityActivation    identity;
SigmoidActivation     sigmoid;
TanhActivation        tanh;
ExponentialActivation exponential;

}  // namespace activation
